@@ -3,8 +3,7 @@
 
 # most of it copied from AWX's scan_packages module
 
-from __future__ import absolute_import, division, print_function
-__metaclass__ = type
+from __future__ import annotations
 
 
 DOCUMENTATION = '''
